from .text_transformer import TextTransformer, TransformerError
from .remove_character_transformer import RemoveCharacter, RemoveCharacterTransformer
from .substitute_character_transformer import (
    CharacterSubstitution,
    SubstituteCharacterTransformer,
)
from .substitute_multiple_characters_transform import (
    MultipleCharacterSubstitution,
    SubstituteMultipleCharactersTransformer,
)
from .consistent_bracketing_transformer import BracketingPair, ConsistentBracketingTransformer



import pandas as pd
from lxml import etree
from dotmap import DotMap
from pathlib import Path
from typing import NamedTuple
from shared_classes import TextChunk, LetterList
from alignment_algorithms import TextAlignments
from text_loader import TextLoader
import numpy as np


class AltoInputPosition(NamedTuple):
    line_id: str
    word_index: int


class AltoAlignmentMap(NamedTuple):
    input_position: AltoInputPosition
    output_idx: int


class AltoXMLTextLoader(TextLoader):
    def __init__(
        self, file_paths: list[Path], estimated_number_of_lines_per_page: int = 60
    ):
        self.__serialized_xml = pd.DataFrame()
        self.__estimated_number_of_lines_per_page = estimated_number_of_lines_per_page
        super().__init__(file_paths)
        self._input_output_map: list[AltoAlignmentMap] = []

    def _load(self) -> LetterList:
        self.import_ocr_texts()
        tokens: list[int] = []
        token_idx_count = 0
        for _, row in self.__serialized_xml.iterrows():
            chunk_start_idx = token_idx_count
            for token_idx, token in enumerate(row.txt):
                token_idx_count += 1
                tokens.append(ord(token))
                self._input_output_map.append(
                    AltoAlignmentMap(
                        AltoInputPosition(str(row.line_ID), token_idx), len(tokens) - 1
                    )
                )
            chunk_end_idx = token_idx_count
            if chunk_start_idx < chunk_end_idx:
                self._text_chunk_indices.append(
                    TextChunk(chunk_start_idx, chunk_end_idx)
                )
        self._output: LetterList = np.array(tokens)

        return super()._load()

    @staticmethod
    def __poly_from_coords(x: int, y: int, w: int, h: int):
        return f"{x} {y} {x + w} {y} {x + w} {y + h} {x} {y + h} {x} {y}"

    def import_ocr_texts(self):
        self._input_output_map = []
        self.__serialized_xml = pd.DataFrame(
            {
                "order": pd.Series([], dtype=pd.UInt16Dtype()),
                "img": pd.Series([], dtype=pd.UInt16Dtype()),
                "zone": pd.Series([], dtype=pd.Float64Dtype()),
                "zone_on_page": pd.Series([], dtype=pd.UInt16Dtype()),
                "line": pd.Series([], dtype=pd.UInt16Dtype()),
                "y_region": pd.Series([], dtype=pd.UInt16Dtype()),
                "x_region": pd.Series([], dtype=pd.UInt16Dtype()),
                "h_region": pd.Series([], dtype=pd.UInt16Dtype()),
                "w_region": pd.Series([], dtype=pd.UInt16Dtype()),
                "y": pd.Series([], dtype=pd.UInt16Dtype()),
                "x": pd.Series([], dtype=pd.UInt16Dtype()),
                "h": pd.Series([], dtype=pd.UInt16Dtype()),
                "w": pd.Series([], dtype=pd.UInt16Dtype()),
                "img_name": pd.Series([], dtype=pd.StringDtype()),
                "region_ID": pd.Series([], dtype=pd.StringDtype()),
                "line_ID": pd.Series([], dtype=pd.StringDtype()),
                "baseline": pd.Series([], dtype=pd.StringDtype()),
                "polygon": pd.Series([], dtype=pd.StringDtype()),
                "txt": pd.Series([], dtype=pd.StringDtype()),
                "aligned_txt": pd.Series([], dtype=pd.StringDtype()),
                "cumul": pd.Series([], dtype=pd.Float64Dtype()),
                "length": pd.Series([], dtype=pd.UInt16Dtype()),
                "imgHeight": pd.Series([], dtype=pd.UInt16Dtype()),
                "imgWidth": pd.Series([], dtype=pd.UInt16Dtype()),
                "regionpolygon": pd.Series([], dtype=pd.StringDtype()),
                "regiontype": pd.Series([], dtype=pd.StringDtype()),
                "regiontagID": pd.Series([], dtype=pd.StringDtype()),
                "linetype": pd.Series([], dtype=pd.StringDtype()),
                "linetagID": pd.Series([], dtype=pd.StringDtype()),
                "xml_filename": pd.Series([], dtype=pd.StringDtype()),
            },
            index=range(
                len(self._file_paths) * self.__estimated_number_of_lines_per_page
            ),
        )  # Start with a reasonably sized table
        order = 0
        image_count = 0
        cumulative_text_length = 0

        for xml_file_path in self._file_paths:
            with open(xml_file_path, "rb") as xml_data:
                events = ("start", "end")
                needed_elements = DotMap(
                    {
                        "file_name": "{http://www.loc.gov/standards/alto/ns-v4#}fileName",
                        "page": "{http://www.loc.gov/standards/alto/ns-v4#}Page",
                        "description": "{http://www.loc.gov/standards/alto/ns-v4#}Description",
                        "other_tag": "{http://www.loc.gov/standards/alto/ns-v4#}OtherTag",
                        "text_block": "{http://www.loc.gov/standards/alto/ns-v4#}TextBlock",
                        "polygon": "{http://www.loc.gov/standards/alto/ns-v4#}Polygon",
                        "text_line": "{http://www.loc.gov/standards/alto/ns-v4#}TextLine",
                        "string": "{http://www.loc.gov/standards/alto/ns-v4#}String",
                    }
                )
                image_filename = ""
                image_height = 0
                image_width = 0
                tag_dict = {}

                zone_on_page_count = 0
                region_id = ""
                region_type_id = ""
                x_region = 0
                y_region = 0
                w_region = 0
                h_region = 0
                region_polygon = ""
                last_parent_tag = ""

                line_in_zone_count = 0
                line_id = ""
                line_type_id = ""
                x = 0
                y = 0
                w = 0
                h = 0
                line_polygon = ""
                line_baseline = ""
                line_text = ""

                for action, element in etree.iterparse(
                    xml_data, events=events, tag=needed_elements.values()
                ):
                    if action == "start":
                        if element.tag == needed_elements.text_line:
                            line_in_zone_count += 1
                            line_id = element.attrib["ID"]
                            line_type_id = element.attrib.get("TAGREFS", "")
                            y = int(float(element.attrib.get("VPOS", 0)))
                            x = int(float(element.attrib.get("HPOS", 0)))
                            h = int(float(element.attrib.get("HEIGHT", 0)))
                            w = int(float(element.attrib.get("WIDTH", 0)))
                            line_baseline = element.attrib.get("BASELINE", "")
                            last_parent_tag = element.tag

                        elif element.tag == needed_elements.string:
                            line_text = element.attrib.get("CONTENT", "")

                        elif element.tag == needed_elements.polygon:
                            if last_parent_tag == needed_elements.text_line:
                                line_polygon = element.attrib.get(
                                    "POINTS", self.__poly_from_coords(x, y, w, h)
                                )
                            elif last_parent_tag == needed_elements.text_block:
                                region_polygon = element.attrib.get(
                                    "POINTS",
                                    self.__poly_from_coords(
                                        x_region, y_region, w_region, h_region
                                    ),
                                )

                        elif element.tag == needed_elements.text_block:
                            zone_on_page_count += 1
                            line_in_zone_count = 0
                            region_id = element.attrib["ID"]
                            region_type_id = element.attrib.get("TAGREFS", "")
                            y_region = int(float(element.attrib.get("VPOS", 0)))
                            x_region = int(float(element.attrib.get("HPOS", 0)))
                            h_region = int(float(element.attrib.get("HEIGHT", 0)))
                            w_region = int(float(element.attrib.get("WIDTH", 0)))
                            last_parent_tag = element.tag

                        elif element.tag == needed_elements.other_tag:
                            if "ID" in element.attrib:
                                tag_dict[element.attrib["ID"]] = element.attrib.get(
                                    "LABEL", ""
                                )

                        elif element.tag == needed_elements.page:
                            image_height = int(float(element.attrib.get("HEIGHT", 0)))
                            image_width = int(float(element.attrib.get("WIDTH", 0)))

                        elif element.tag == needed_elements.file_name:
                            image_count += 1
                            image_filename = element.text

                    if action == "end":
                        if element.tag == needed_elements.text_line:
                            line_length = len(line_text)
                            cumulative_text_length += line_length
                            new_row = [
                                order,
                                image_count,
                                int(f"{image_count}0{zone_on_page_count:02}"),
                                zone_on_page_count,
                                line_in_zone_count,
                                y_region,
                                x_region,
                                h_region,
                                w_region,
                                y,
                                x,
                                h,
                                w,
                                image_filename,
                                region_id,
                                line_id,
                                line_baseline,
                                line_polygon,
                                line_text,
                                "".join([" " for x in line_text]),
                                cumulative_text_length,
                                line_length,
                                image_height,
                                image_width,
                                region_polygon,
                                tag_dict.get(region_type_id, ""),
                                region_type_id,
                                tag_dict.get(line_type_id, ""),
                                line_type_id,
                                xml_file_path.name,
                            ]
                            if order < len(self.__serialized_xml):
                                self.__serialized_xml.at[
                                    order, self.__serialized_xml.columns
                                ] = new_row
                            else:
                                self.__serialized_xml.append(
                                    pd.DataFrame(
                                        [new_row], columns=self.__serialized_xml.columns
                                    )
                                )

                            order += 1
                        element.clear()

        # Crop to only rows with values
        self.__serialized_xml = self.__serialized_xml.loc[
            self.__serialized_xml.first_valid_index() : self.__serialized_xml.last_valid_index()
        ]
        self.__serialized_xml["img_name"] = self.__serialized_xml.img_name.astype(
            "category"
        )
        self.__serialized_xml["region_ID"] = self.__serialized_xml.region_ID.astype(
            "category"
        )
        self.__serialized_xml["line_ID"] = self.__serialized_xml.line_ID.astype(
            "category"
        )
        self.__serialized_xml["baseline"] = self.__serialized_xml.baseline.astype(
            "category"
        )
        self.__serialized_xml["polygon"] = self.__serialized_xml.polygon.astype(
            "category"
        )
        self.__serialized_xml["txt"] = self.__serialized_xml.txt.astype("category")
        self.__serialized_xml["aligned_txt"] = self.__serialized_xml.aligned_txt.astype(
            pd.StringDtype()
        )
        self.__serialized_xml[
            "regionpolygon"
        ] = self.__serialized_xml.regionpolygon.astype("category")
        self.__serialized_xml["regiontype"] = self.__serialized_xml.regiontype.astype(
            "category"
        )
        self.__serialized_xml["regiontagID"] = self.__serialized_xml.regiontagID.astype(
            "category"
        )
        self.__serialized_xml["linetype"] = self.__serialized_xml.linetype.astype(
            "category"
        )
        self.__serialized_xml["linetagID"] = self.__serialized_xml.linetagID.astype(
            "category"
        )
        self.__serialized_xml[
            "xml_filename"
        ] = self.__serialized_xml.xml_filename.astype("category")

    def emend_original_with_aligned_text(
        self, alignment_pairs: TextAlignments, aligned_text: LetterList
    ) -> pd.DataFrame:
        # TODO: the emandation is not working quite right.
        alignment_pairs_dict: dict[int, list[int]] = {}
        for alignment in alignment_pairs.alignments:
            if alignment.query_idx not in alignment_pairs_dict:
                alignment_pairs_dict[alignment.query_idx] = []

            alignment_pairs_dict[alignment.query_idx].append(alignment.target_idx)

        input_to_aligned_text_map: list[AltoAlignmentMap] = []
        for input_position, output_idx in [
            (x.input_position, x.output_idx) for x in self._input_output_map
        ]:
            if output_idx in alignment_pairs_dict:
                for align_idx in alignment_pairs_dict[output_idx]:
                    input_to_aligned_text_map.append(
                        AltoAlignmentMap(input_position, align_idx)
                    )
                    pass

        input_to_aligned_text_map_dict: dict[str, list[int]] = {}
        # Yeah, that's a hacky bunch of shit in the middle to make the input key unique (it's almost the weekend and I don't want to think about the "right" way to do it)
        garbage = "@#$%^&"
        for alignment in input_to_aligned_text_map:
            unique_input_key = f"{alignment.input_position.line_id}{garbage}{alignment.input_position.word_index}"
            if unique_input_key not in alignment_pairs_dict:
                input_to_aligned_text_map_dict[unique_input_key] = []

            input_to_aligned_text_map_dict[unique_input_key].append(
                alignment.output_idx
            )

        # aligned_transcription: list[str] = []
        used_alinment_text_indices: list[int] = []
        for idx, entry in self.__serialized_xml.iterrows():
            line_text = ""
            for char_idx in range(len(entry.txt)):
                unique_input_key = f"{entry['line_ID']}{garbage}{char_idx}"
                if unique_input_key not in input_to_aligned_text_map_dict:
                    line_text += entry.txt[char_idx]
                    continue

                for alignment_idx in input_to_aligned_text_map_dict[unique_input_key]:
                    if alignment_idx not in used_alinment_text_indices:
                        used_alinment_text_indices.append(alignment_idx)
                        line_text += chr(aligned_text[alignment_idx])
            self.__serialized_xml.at[idx, "aligned_txt"] = line_text

        return self.__serialized_xml

    def replace_original_with_aligned_text(
        self, alignment_pairs: TextAlignments, aligned_text: LetterList
    ) -> pd.DataFrame:
        # TODO: as written this was originally going to be used to correct individual letters in the original
        # I repurposed it to simply replace the text with its match, but that can probably be done more simply.
        alignment_pairs_dict: dict[int, list[int]] = {}
        for alignment in alignment_pairs.alignments:
            if alignment.query_idx not in alignment_pairs_dict:
                alignment_pairs_dict[alignment.query_idx] = []

            alignment_pairs_dict[alignment.query_idx].append(alignment.target_idx)

        input_to_aligned_text_map: list[AltoAlignmentMap] = []
        for input_position, output_idx in [
            (x.input_position, x.output_idx) for x in self._input_output_map
        ]:
            if output_idx in alignment_pairs_dict:
                for align_idx in alignment_pairs_dict[output_idx]:
                    input_to_aligned_text_map.append(
                        AltoAlignmentMap(input_position, align_idx)
                    )
                    pass

        input_to_aligned_text_map_dict: dict[str, list[int]] = {}
        # Yeah, that's a hacky bunch of shit in the middle to make the input key unique (it's almost the weekend and I don't want to think about the "right" way to do it)
        garbage = "@#$%^&"
        for alignment in input_to_aligned_text_map:
            unique_input_key = f"{alignment.input_position.line_id}{garbage}{alignment.input_position.word_index}"
            if unique_input_key not in alignment_pairs_dict:
                input_to_aligned_text_map_dict[unique_input_key] = []

            input_to_aligned_text_map_dict[unique_input_key].append(
                alignment.output_idx
            )

        # aligned_transcription: list[str] = []
        used_alinment_text_indices: list[int] = []
        for idx, entry in self.__serialized_xml.iterrows():
            line_text = ""
            for char_idx in range(len(entry.txt)):
                unique_input_key = f"{entry['line_ID']}{garbage}{char_idx}"
                # if unique_input_key not in input_to_aligned_text_map_dict:
                #     line_text += entry.txt[char_idx]
                #     continue
                if unique_input_key in input_to_aligned_text_map_dict:
                    for alignment_idx in input_to_aligned_text_map_dict[
                        unique_input_key
                    ]:
                        if alignment_idx not in used_alinment_text_indices:
                            used_alinment_text_indices.append(alignment_idx)
                            line_text += chr(aligned_text[alignment_idx])
            self.__serialized_xml.at[idx, "aligned_txt"] = line_text

        return self.__serialized_xml



from typing import Union, Any
from shared_classes import LetterList, TextChunk
from alignment_tool import compare_parallel_text_chunks
import numpy as np


class AlignmentException(Exception):
    pass


class AlignmentKey:
    def __init__(self, query_idx: int, target_idx: int):
        self.query_idx = query_idx
        self.target_idx = target_idx


class TextAlignments:
    def __init__(self):
        self.alignments: list[AlignmentKey] = []


class AlignmentAlgorithm:
    """
    A class for performing a single text alignment operation.
    It may be used to redefine text chunks as well.
    """

    def __init__(self):
        pass

    def load_texts(
        self,
        query: LetterList,
        target: LetterList,
        input_query_text_chunk_indices: list[TextChunk],
        input_target_text_chunk_indices: list[TextChunk],
    ):
        self._query = query
        self._target = target
        self._input_query_text_chunk_indices = input_query_text_chunk_indices
        self._input_target_text_chunk_indices = input_target_text_chunk_indices
        self._output_query_text_chunk_indices: list[TextChunk] = []
        self._output_target_text_chunk_indices: list[TextChunk] = []
        self._alignment: TextAlignments = TextAlignments()

    def align(self) -> TextAlignments:
        if self._alignment.alignments:
            return self._alignment

        if self._query.size == 0:
            raise AlignmentException("The query text is empty")
        if self._target.size == 0:
            raise AlignmentException("The target text is empty")

        return TextAlignments()

    def apply_alignment(
        self, alignment_pairs: TextAlignments, aligned_text: LetterList
    ) -> Union[LetterList, Any]:
        input_to_aligned_text_map = TextAlignments()
        alignment_pairs_dict: dict[int, list[int]] = {}

        for alignment in alignment_pairs.alignments:
            if alignment.query_idx not in alignment_pairs_dict:
                alignment_pairs_dict[alignment.query_idx] = []

            alignment_pairs_dict[alignment.query_idx].append(alignment.target_idx)

        for input_idx, output_idx in [
            (x.query_idx, x.target_idx) for x in self.input_output_alignment.alignments
        ]:
            if output_idx in alignment_pairs_dict:
                for align_idx in alignment_pairs_dict[output_idx]:
                    input_to_aligned_text_map.alignments.append(
                        AlignmentKey(input_idx, align_idx)
                    )
                    pass

        input_to_aligned_text_map_dict: dict[int, list[int]] = {}
        for alignment in input_to_aligned_text_map.alignments:
            if alignment.query_idx not in alignment_pairs_dict:
                input_to_aligned_text_map_dict[alignment.query_idx] = []

            input_to_aligned_text_map_dict[alignment.query_idx].append(
                alignment.target_idx
            )

        input_with_alignment_applied = []
        for input_idx in range(self._query.size):
            if input_idx not in input_to_aligned_text_map_dict:
                input_with_alignment_applied.append(self._query[input_idx])
                continue

            for alignment_idx in input_to_aligned_text_map_dict[input_idx]:
                input_with_alignment_applied.append(aligned_text[alignment_idx])

        return np.array(input_with_alignment_applied)

    @property
    def chunked_query_text(self) -> list[tuple[str, str]]:
        """Retrieve a visualization of the current chunked query text

        Returns:
            list[tuple[str, str]]: A list containing the name and transcription of each text chunk
        """

        return [
            (y.name, "".join([chr(self._query[x]) for x in y.indices]))
            for y in self.output_query_text_chunk_indices
        ]

    @property
    def chunked_target_text(self) -> list[tuple[str, str]]:
        """Retrieve a visualization of the current chunked target text

        Returns:
            list[tuple[str, str]]: A list containing the name and transcription of each text chunk
        """

        return [
            (y.name, "".join([chr(self._target[x]) for x in y.indices]))
            for y in self.output_target_text_chunk_indices
        ]

    @property
    def aligned_text(self, rtl=False) -> str:
        """Visualize the alignment of the two texts

        Args:
            rtl (bool, optional): Set the text direction to right-to-left. Defaults to False.

        Returns:
            str: A tabular representation of the text alignment
        """

        return compare_parallel_text_chunks(
            self._query,
            self.output_query_text_chunk_indices,
            self._target,
            self.output_target_text_chunk_indices,
            rtl,
        )

    @property
    def input_output_alignment(self) -> TextAlignments:
        return self._alignment

    @property
    def input_query_text_chunk_indices(self) -> list[TextChunk]:
        if not self._input_query_text_chunk_indices:
            self._input_query_text_chunk_indices.append(
                TextChunk([x for x in range(0, self._query.size)], "default full text")
            )

        if self._input_query_text_chunk_indices:
            return self._input_query_text_chunk_indices

        raise AlignmentException(
            "No input query text chunk was created (this is a development error)"
        )

    @property
    def input_target_text_chunk_indices(self) -> list[TextChunk]:
        if not self._input_target_text_chunk_indices:
            self._input_target_text_chunk_indices.append(
                TextChunk([x for x in range(0, self._target.size)], "default full text")
            )

        if self._input_target_text_chunk_indices:
            return self._input_target_text_chunk_indices

        raise AlignmentException(
            "No input target text chunk was created (this is a development error)"
        )

    @property
    def output_query_text_chunk_indices(self) -> list[TextChunk]:
        if not self._output_query_text_chunk_indices:
            return self.input_query_text_chunk_indices

        return self._output_query_text_chunk_indices

    @property
    def output_target_text_chunk_indices(self) -> list[TextChunk]:
        if not self._output_target_text_chunk_indices:
            self.input_target_text_chunk_indices

        return self._output_target_text_chunk_indices



from shared_classes import TextChunk, LetterList
from text_loader import TextLoader
import numpy as np


class StringTextLoader(TextLoader):
    def __init__(self, input_text: str):
        super().__init__()
        self._output = np.array([ord(x) for x in input_text])
        self._text_chunk_indices = [
            TextChunk(list(range(0, self.output.size)), "full text")
        ]
        self._input_output_map = [(x, x) for x in range(0, self.output.size)]

    def _load(self) -> LetterList:
        return super()._load()



from shared_classes import LetterList, TextChunk
from text_transformers import TextTransformer
import numpy as np


class CharacterSubstitution:
    def __init__(self, input_character: str, output_character: str) -> None:
        self.input_character = self.verify_character(input_character)
        self.output_character = self.verify_character(output_character)

    @staticmethod
    def verify_character(char: str):
        if not isinstance(char, (str)):
            raise TypeError(f"Only str types are accepted: {char} is {type(char)} ")
        if len(char) != 1:
            raise Exception(
                f"Only single characters are accepted, but {char} with length {len(char)} was submitted."
            )
        return char


class SubstituteCharacterTransformer(TextTransformer):
    def __init__(self, substitution_list: list[CharacterSubstitution]):
        """Perform a transform of a single character to another single character

        Args:
            substitution_list (list[CharacterSubstitution]): a list of input/output character substitutions
        """
        self.substitution_list = substitution_list
        super().__init__()

    def load_input(self, input: LetterList, text_chunk_indices: list[TextChunk]):
        super().load_input(input, text_chunk_indices, False)

    def transform(self) -> LetterList:
        if np.any(self._output):
            return self._output

        self.__transform()

        return super().transform()

    def __transform(self):
        replacement_codes = {
            ord(x.input_character): ord(x.output_character)
            for x in self.substitution_list
        }

        self._input_output_map = [(idx, idx) for idx in range(len(self.input))]
        self._output = np.vectorize(replacement_codes.get)(self.input, self.input)
        self._text_chunk_indices = self._input_text_chunk_indices



from shared_classes import LetterList, TextChunk
from text_transformers import TextTransformer
import numpy as np


class RemoveCharacter:
    def __init__(self, remove_character: str) -> None:
        self.remove_character = self.verify_character(remove_character)

    @staticmethod
    def verify_character(char: str):
        if not isinstance(char, (str)):
            raise TypeError(f"Only str types are accepted: {char} is {type(char)} ")
        if len(char) != 1:
            raise Exception(
                f"Only single characters are accepted, but {char} with length {len(char)} was submitted."
            )
        return char


class RemoveCharacterTransformer(TextTransformer):
    def __init__(self, remove_character_list: list[RemoveCharacter]):
        """Remove every instance of the submitted characters

        Args:
            substitution_list (list[CharacterSubstitution]): a list of characters to be removed
        """
        self.__remove_character_list = remove_character_list
        super().__init__()

    def load_input(self, input: LetterList, text_chunk_indices: list[TextChunk]):
        super().load_input(input, text_chunk_indices, False)

    def transform(self) -> LetterList:
        if np.any(self._output):
            return self._output

        self.__transform()

        return super().transform()

    def __transform(self):
        remove_characters = set(
            [ord(x.remove_character) for x in self.__remove_character_list]
        )

        output: list[int] = []
        output_chunks: list[TextChunk] = []
        input_output_map: list[tuple[int, int]] = []
        for chunk in self._input_text_chunk_indices:
            output_chunk = TextChunk([], chunk.name)
            output_chunk_start_idx = len(output)
            chunk_text = self.input[chunk.indices]
            for idx, token in enumerate(chunk_text):
                if token in remove_characters:
                    continue  # Just don't add the token to output to "remove" it
                current_idx = chunk.indices[0] + idx
                output.append(token)
                output_chunk.indices.append(len(output))
                input_output_map.append((current_idx, len(output) - 1))

            if output_chunk_start_idx <= len(output) - 1:
                output_chunks.append(output_chunk)

        self._input_output_map = input_output_map
        self._output = np.array(output, dtype=np.uint32)
        self._text_chunk_indices = output_chunks



import logging
import edlib
from .alignment_algorithm import AlignmentAlgorithm, TextAlignments, AlignmentKey
from ..shared_classes import LetterList, TextChunk


class GlobalAlignmentAlgorithm(AlignmentAlgorithm):
    def __init__(self):
        super().__init__()

    def align(self) -> TextAlignments:
        if self._alignment.alignments:
            return self._alignment

        self._alignment = global_alignment(self._query, self._target, self._input_query_text_chunk_indices, self._input_target_text_chunk_indices)
        return self._alignment


def global_alignment(query_text: LetterList, target_text: LetterList, query_text_chunks: list[TextChunk], target_text_chunks: list[TextChunk]) -> TextAlignments:
    alignments = TextAlignments()

    if len(query_text_chunks) != len(target_text_chunks):
        logging.warning(f'Since the query has {len(query_text_chunks)} text chunks and the target has {len(target_text_chunks)}, the alignment will probably not work as intended.')

    query_start_end = [(x.start_idx, x.end_idx) for x in query_text_chunks]
    target_start_end = [(x.start_idx, x.end_idx) for x in target_text_chunks]
    for (query_start_idx, query_end_idx), (target_start_idx, target_end_idx) in zip(query_start_end, target_start_end):
    # for query_text_chunk, target_text_chunk in zip(chunked_query_text, chunked_target_text):
        query_text_chunk = query_text[query_start_idx:query_end_idx]
        target_text_chunk = target_text[target_start_idx:target_end_idx]
        query = ''.join([chr(x) for x in query_text_chunk])
        target = ''.join([chr(x) for x in target_text_chunk])

        result = edlib.align(query, target, task = "path", mode='HW')
        nice = edlib.getNiceAlignment(result, query, target)
        query_count_idx = query_start_idx
        target_count_idx = target_start_idx
        for idx in range(len(nice['query_aligned'])):
            alignments.alignments.append(AlignmentKey(query_count_idx,target_count_idx))
            if nice['query_aligned'][idx] != '-':
                query_count_idx += 1
            if nice['target_aligned'][idx] != '-':
                target_count_idx += 1        

    return alignments



from aligner import (
    TextAlignmentException,
    AlignedIndices,
    AlignmentTextDataObject,
    AlignmentOperation,
)
from alignment_tool import (
    TextAlignmentTool,
    get_text_for_range,
    get_text_chunk_string,
    get_text_chunks_string,
    compare_parallel_text_chunks,
    print_parallel_text_chunks,
)



from itertools import groupby
from typing import Counter, Union
import numpy as np
import re
import bidict


# See https://stackoverflow.com/questions/4998629/split-string-with-multiple-delimiters-in-python
def split(delimiters: list[str], string: str, maxsplit: int = 0) -> list[str]:
    regexPattern = "|".join(map(re.escape, delimiters))
    return re.split(regexPattern, string, maxsplit)


def letter_list_to_unique_word_codes(
    letter_list: np.ndarray, separators: Union[list[str], None] = None
) -> tuple[np.ndarray, bidict.bidict[str, int], dict[int, list[int]]]:
    """
    This assigns a unique number to each unique word in the submitted
    text. It returns an array with the text represented by its unique
    word codes along with the word_code_key.
    """

    if separators is None:
        separators = [" "]

    indexed_text = [(chr(x), idx) for idx, x in enumerate(letter_list)]
    word_chunks = (
        list(g)
        for k, g in groupby(indexed_text, key=lambda x: x[0] not in separators)
        if k
    )
    word_map: dict[str, int] = {}
    coded_output: list[int] = []
    word_to_letter_map: dict[int, list[int]] = {}
    current_word_idx = 0
    for idx, word in enumerate(word_chunks):
        text_word = "".join([x[0] for x in word])
        if text_word not in word_map:
            word_map[text_word] = current_word_idx
            current_word_idx = current_word_idx + 1
        coded_output.append(word_map[text_word])
        word_to_letter_map[idx] = [x[1] for x in word]

    return (
        np.array(coded_output).astype(np.uint32),
        bidict.bidict(word_map),
        word_to_letter_map,
    )


def find_wordlist_for_alignment(
    input_gold_text: np.ndarray, max_word_frequency: int = 100
) -> list[int]:
    gold_text_map = [
        x[0] for x in Counter(input_gold_text).items() if x[1] < max_word_frequency
    ]
    return list(set(gold_text_map))


def filter_wordlist(
    wordlist: np.ndarray, valid_values: list[int]
) -> tuple[np.ndarray, dict[int, int]]:
    filtered_list = [(idx, x) for idx, x in enumerate(wordlist) if x in valid_values]
    return np.array([x[1] for x in filtered_list]).astype(np.uint32), {
        x[0]: idx for idx, x in enumerate(filtered_list)
    }



from alignment_algorithm import (
    AlignmentAlgorithm,
    TextAlignments,
    AlignmentKey,
    AlignmentException,
)
from global_alignment import GlobalAlignmentAlgorithm
from local_alignment import LocalAlignmentAlgorithm
from rough_alignment import RoughAlignmentAlgorithm
from chunk_alignment import ChunkAlignmentAlgorithm



from shared_classes import LetterList, TextChunk
from text_transformers import TextTransformer
import numpy as np


class BracketingPair:
    def __init__(self, first_token: str, second_token: str) -> None:
        self.__pair = [
            self.verify_character(first_token),
            self.verify_character(second_token),
        ]
        self.__original_order = [ord(first_token), ord(second_token)]

    @staticmethod
    def verify_character(char: str):
        if not isinstance(char, (str)):
            raise TypeError(f"Only str types are accepted: {char} is {type(char)} ")
        if len(char) != 1:
            raise Exception(
                f"Only single characters are accepted, but {char} with length {len(char)} was submitted."
            )
        return ord(char)

    def __getitem__(self, i: int):
        return self.__pair[i]

    def __len__(self):
        return len(self.__pair)

    def __eq__(self, other: int):
        return other in self.__pair

    def reverse(self):
        self.__pair.reverse()

    def reset(self):
        self.__pair[0] = self.__original_order[0]
        self.__pair[1] = self.__original_order[1]


class RemoveCharacter:
    def __init__(self, remove_character: str) -> None:
        self.remove_character = self.verify_character(remove_character)

    @staticmethod
    def verify_character(char: str):
        if not isinstance(char, (str)):
            raise TypeError(f"Only str types are accepted: {char} is {type(char)} ")
        if len(char) != 1:
            raise Exception(
                f"Only single characters are accepted, but {char} with length {len(char)} was submitted."
            )
        return char


class ConsistentBracketingTransformer(TextTransformer):
    def __init__(self, bracketing_pair_list: list[BracketingPair]):
        """Look for pairs of bracketing characters and ensure they are
        paired consistently.

        Args:
            substitution_list (list[CharacterSubstitution]): a list of characters to be removed
        """
        self.__bracketing_pair_list = bracketing_pair_list
        super().__init__()

    def load_input(self, input: LetterList, text_chunk_indices: list[TextChunk]):
        super().load_input(input, text_chunk_indices, False)

    def transform(self) -> LetterList:
        if np.any(self._output):
            return self._output

        self.__transform()

        return super().transform()

    def __transform(self):
        output: list[int] = []
        for chunk in self._input_text_chunk_indices:
            chunk_text = self.input[chunk.indices]
            for token in chunk_text:
                for bracketing_pair in self.__bracketing_pair_list:
                    if token in bracketing_pair:
                        if token != bracketing_pair[0]:
                            token = bracketing_pair[0]
                        bracketing_pair.reverse()
                output.append(token)
            for bracketing_pair in self.__bracketing_pair_list:
                bracketing_pair.reset()

        self._output = np.array(output, dtype=np.uint32)
        self._input_output_map = [(idx, idx) for idx in range(self._output.size)]
        self._text_chunk_indices = self._input_text_chunk_indices



from .alignment_algorithm import AlignmentAlgorithm, TextAlignments
import swalign
# from ..alignment_tool import AlignmentTextPair, AlignmentRange
from ..find_wordlist_for_alignment import find_wordlist_for_alignment


class RoughAlignmentAlgorithm(AlignmentAlgorithm):
    def __init__(self):
        super().__init__()

    def align(self) -> TextAlignments:

        query_text = self._query
        target_text = self._target
        text_hash = {x: idx for idx, x in enumerate(list(set(query_text + target_text)))}
        less_common_text_words = find_wordlist_for_alignment(target_text, 100)

        unique_target_text_matching_word_codes = {}
        query_text_token_codes = []
        for line_element in query_text:
            if line_element not in less_common_text_words:
                continue

            target_text_matches = [idx for idx, x in enumerate(target_text) if x == line_element]
            word_code = text_hash[line_element]
            query_text_token_codes.append(word_code)
            for target_text_match in target_text_matches:
                unique_target_text_matching_word_codes[f'{target_text_match}_{word_code}'] = {'index': target_text_match, 'wordcode': word_code}

        # Get matching words as a list sorted in order of occurrence
        sorted_unique_target_text_matching_word_codes = sorted(
            list(unique_target_text_matching_word_codes.values()), key=lambda x: x['index'])
        if not sorted_unique_target_text_matching_word_codes:
            return alignment_data
                
        rough_start_in_matching_word_codes, rough_end_in_matching_word_codes = rough_alignment(
            target_text, sorted_unique_target_text_matching_word_codes, query_text_token_codes, text_hash)

        rough_start_in_matching_word_codes = rough_start_in_matching_word_codes if rough_start_in_matching_word_codes >= len(sorted_unique_target_text_matching_word_codes) else len(sorted_unique_target_text_matching_word_codes) -1
        rough_start_in_matching_word_codes = rough_start_in_matching_word_codes if rough_start_in_matching_word_codes < 0 else 0
        rough_end_in_matching_word_codes = rough_end_in_matching_word_codes if rough_end_in_matching_word_codes >= len(sorted_unique_target_text_matching_word_codes) else len(sorted_unique_target_text_matching_word_codes) -1
        rough_end_in_matching_word_codes = rough_end_in_matching_word_codes if rough_end_in_matching_word_codes < 0 else 0

        rough_start_in_target_text = sorted_unique_target_text_matching_word_codes[rough_start_in_matching_word_codes]['index']
        rough_end_in_target_text = sorted_unique_target_text_matching_word_codes[rough_end_in_matching_word_codes]['index']

        query_text_length = len(query_text)

        # Take the rough match and extend the search outward by half the length of the query string on each side
        padded_match_start = max(0, int(rough_start_in_target_text - (query_text_length / 2)))
        padded_match_end = min(len(target_text), int(rough_end_in_target_text + (query_text_length / 2)))
        alignment_data.alignment_range.append(
            AlignmentRange(
                query_start=0,
                query_end=len(query_text) - 1,
                target_start=padded_match_start,
                target_end=padded_match_end))

        
        return super().align()


def rough_alignment(target_text, unique_target_text_matching_word_codes, query_word_codes, text_hash):
    alignment_word_codes = [text_hash[x] for x in target_text if x in unique_target_text_matching_word_codes]

    # choose your own values here… 2 and -1 are common.
    match = 2
    mismatch = -1
    scoring = swalign.NucleotideScoringMatrix(match, mismatch)

    sw = swalign.LocalAlignment(scoring)  # you can also choose gap penalties, etc...
    target_text_symbolic_transcription = ''.join([chr(x) for x in alignment_word_codes])
    query_symbolic_transcription = ''.join([chr(x) for x in query_word_codes])
    alignment = sw.align(target_text_symbolic_transcription, query_symbolic_transcription)
    start_idx, end_idx = alignment.r_pos, alignment.r_end
    # alignment.dump()
    return start_idx, end_idx


# def align(alignment_data: AlignmentTextPair):
#     less_common_text_words = find_wordlist_for_alignment(alignment_data.target.text, 100)
#     query_text = alignment_data.query.text
#     target_text = alignment_data.target.text

#     zones = alignment_data.query.text.zone.unique()
#     for zone in zones:
#         query_zone = query_text[query_text.zone == zone]
#         unique_target_text_matching_word_codes = {}
#         query_zone_word_codes = []
#         for line_element in itertools.chain(*[HTR_line_text.split() for HTR_line_text in query_zone['txt']]):
#             if line_element not in less_common_text_words:
#                 continue

#             target_text_matches = target_text[target_text.text == line_element]
#             word_code = target_text_matches.iloc[0, :].cat.codes.text
#             query_zone_word_codes.append(word_code)
#             for target_text_match in target_text_matches['text'].index:
#                 unique_target_text_matching_word_codes[f'{target_text_match}_{word_code}'] = {'index': target_text_match, 'wordcode': word_code}

#         # Get matching words as a list sorted in order of occurrence
#         sorted_unique_target_text_matching_word_codes = sorted(
#             list(unique_target_text_matching_word_codes.values()), key=lambda x: x['index'])
#         if not sorted_unique_target_text_matching_word_codes:
#             continue
            
#         rough_start_in_matching_word_codes, rough_end_in_matching_word_codes = rough_alignment(
#             target_text, sorted_unique_target_text_matching_word_codes, query_zone_word_codes)

#         rough_start_in_matching_word_codes = rough_start_in_matching_word_codes if rough_start_in_matching_word_codes >= len(sorted_unique_target_text_matching_word_codes) else len(sorted_unique_target_text_matching_word_codes) -1
#         rough_start_in_matching_word_codes = rough_start_in_matching_word_codes if rough_start_in_matching_word_codes < 0 else 0
#         rough_end_in_matching_word_codes = rough_end_in_matching_word_codes if rough_end_in_matching_word_codes >= len(sorted_unique_target_text_matching_word_codes) else len(sorted_unique_target_text_matching_word_codes) -1
#         rough_end_in_matching_word_codes = rough_end_in_matching_word_codes if rough_end_in_matching_word_codes < 0 else 0

#         rough_start_in_target_text = sorted_unique_target_text_matching_word_codes[rough_start_in_matching_word_codes]['index']
#         rough_end_in_target_text = sorted_unique_target_text_matching_word_codes[rough_end_in_matching_word_codes]['index']

#         query_zone_length = len(query_zone)

#         # Take the rough match and extend the search outward by half the length of the query string on each side
#         padded_match_start = max(0, int(rough_start_in_target_text - (query_zone_length / 2)))
#         padded_match_end = min(len(target_text), int(rough_end_in_target_text + (query_zone_length / 2)))
#         alignment_data.alignment_range.append(
#             AlignmentRange(
#                 query_start=query_zone.index[0],
#                 query_end=query_zone.index[-1],
#                 target_start=padded_match_start,
#                 target_end=padded_match_end))

#     return alignment_data


# def rough_alignment(target_text, unique_target_text_matching_word_codes, query_word_codes):
#     alignment_word_codes = list(
#         target_text[target_text.text.cat.codes.isin([x['wordcode'] for x in unique_target_text_matching_word_codes])].text.cat.codes)

#     # choose your own values here… 2 and -1 are common.
#     match = 2
#     mismatch = -1
#     scoring = swalign.NucleotideScoringMatrix(match, mismatch)

#     sw = swalign.LocalAlignment(scoring)  # you can also choose gap penalties, etc...
#     target_text_symbolic_transcription = ''.join([chr(x) for x in alignment_word_codes])
#     query_symbolic_transcription = ''.join([chr(x) for x in query_word_codes])
#     alignment = sw.align(target_text_symbolic_transcription, query_symbolic_transcription)
#     start_idx, end_idx = alignment.r_pos, alignment.r_end
#     # alignment.dump()
#     return start_idx, end_idx



from shared_classes import LetterList, TextChunk
from alignment_algorithms import (
    TextAlignments,
    AlignmentAlgorithm,
)
from text_transformers import TextTransformer
from typing import NamedTuple, Union
from enum import Enum


class TextAlignmentException(Exception):
    pass


class AlignedIndices(NamedTuple):
    input_idx: int
    output_idx: int


class TextType(Enum):
    QUERY = 1
    TARGET = 2


class QueryTargetAlignment:
    def __init__(
        self,
        query_to_target_mapping: TextAlignments,
        query_text: LetterList,
        target_text: LetterList,
        query_text_chunks: list[TextChunk],
        target_text_chunks: list[TextChunk],
        query_transformer_class_name: str,
        target_transformer_class_name: str,
    ):
        self.query_to_target_mapping: TextAlignments = query_to_target_mapping
        self.query_text: LetterList = query_text
        self.target_text: LetterList = target_text
        self.query_text_chunks: list[TextChunk] = query_text_chunks
        self.target_text_chunks: list[TextChunk] = target_text_chunks
        self.__query_transformer_class_name = query_transformer_class_name
        self.__target_transformer_class_name = target_transformer_class_name
        self.__query_to_target_chunk_mapping: list[tuple[TextChunk, TextChunk]] = []

    @property
    def query_to_target_chunk_mapping(self) -> list[tuple[TextChunk, TextChunk]]:
        # TODO: perhaps update all such "cached properties" to @functools.cached_property
        # but support is touchy in python versions: @functools.cached_property >= Python 3.8;
        # @property + @functools.lru_cache() >= Python 3.2, < Python 3.8; ...
        if len(self.__query_to_target_chunk_mapping) > 0:
            return self.__query_to_target_chunk_mapping

        aligned_text_chunks_set: set[tuple[int, int]] = set([])
        aligned_text_chunks: list[tuple[TextChunk, TextChunk]] = []
        alignments_idx = 0
        while alignments_idx is not None and alignments_idx < len(
            self.query_to_target_mapping.alignments
        ):
            current_alignment = self.query_to_target_mapping.alignments[alignments_idx]
            stop_query = False
            for q_idx, query_text_chunk in enumerate(self.query_text_chunks):
                for t_idx, target_text_chunk in enumerate(self.target_text_chunks):
                    if (
                        query_text_chunk.start_idx
                        <= current_alignment.query_idx
                        <= query_text_chunk.end_idx
                        and target_text_chunk.start_idx
                        <= current_alignment.target_idx
                        <= target_text_chunk.end_idx
                    ):
                        if (q_idx, t_idx) not in aligned_text_chunks_set:
                            aligned_text_chunks_set.add((q_idx, t_idx))
                            aligned_text_chunks.append(
                                (query_text_chunk, target_text_chunk)
                            )
                        stop_query = True
                        break
                if stop_query:
                    break
            alignments_idx = alignments_idx + 1
        self.__query_to_target_chunk_mapping = aligned_text_chunks
        return self.__query_to_target_chunk_mapping

    @property
    def unaligned_query_text_chunks(self) -> list[TextChunk]:
        matched_query_chunks = [x[0] for x in self.query_to_target_chunk_mapping]
        return list(
            dict.fromkeys(
                [x for x in self.query_text_chunks if x not in matched_query_chunks]
            )
        )

    @property
    def unaligned_target_text_chunks(self) -> list[TextChunk]:
        matched_target_chunks = [x[1] for x in self.query_to_target_chunk_mapping]
        return list(
            dict.fromkeys(
                [x for x in self.target_text_chunks if x not in matched_target_chunks]
            )
        )

    @property
    def query_transformer_class_name(self):
        return self.__query_transformer_class_name

    @property
    def target_transformer_class_name(self):
        return self.__target_transformer_class_name


class AlignmentTextDataObject:
    def __init__(self):
        self.text: list[str] = []


class AlignmentOperation:
    """
    A container for the specific details of a transform or
    alignment operation.
    """

    def __init__(
        self,
        query_transformation: Union[TextTransformer, None],
        target_transformation: Union[TextTransformer, None],
        alignment: Union[AlignmentAlgorithm, None],
    ):
        self.query_transformation = query_transformation
        self.target_transformation = target_transformation
        self.alignment = alignment



from pathlib import Path
from typing import Any, Union
import numpy as np
from shared_classes import TextChunk, LetterList


class LoaderError(Exception):
    pass


# TODO: the loader should probably be able to recreate the original
# query text, but replacing its text with the aligned target text.
class TextLoader:
    """
    Instructions for converting the input source into
    the data structures used by the alignment tool.
    """

    def __init__(self, file_path: Union[Path, None] = None):
        if file_path is not None:
            self._file_path = file_path
        self._output: np.ndarray = np.array([], dtype=np.uint32)
        self._input_output_map: Any = None
        self._text_chunk_indices: list[TextChunk] = []

    def _load(self) -> LetterList:
        if self._output.size > 0:
            return self._output

        raise LoaderError("Transform action resulted in empty output")

    @property
    def output(self) -> LetterList:
        if self._output.size == 0:
            _ = self._load()

        if self._output.size > 0:
            return self._output

        raise LoaderError("Output is empty")

    @property
    def input_output_map(self) -> Any:
        if not self._input_output_map:
            _ = self.output

        if self._input_output_map:
            return self._input_output_map

        raise LoaderError("Input output map is empty")

    @property
    def text_chunk_indices(self) -> list[TextChunk]:
        if not self._text_chunk_indices:
            self._text_chunk_indices.append(TextChunk([], ""))

        if self._text_chunk_indices:
            return self._text_chunk_indices

        raise LoaderError("No text chunk was created (this is a development error)")



from dataclasses import dataclass
import numpy as np


LetterList = np.ndarray


@dataclass(init=True, frozen=True)
class TextChunk:
    __slots__ = ("indices", "name")
    indices: list[int]
    name: str



import itertools
import sys

from alignment_algorithms import (
    AlignmentKey,
    local_alignment,
)
from typing import Sequence, Union
from alignment_algorithms import AlignmentException, TextAlignments, AlignmentAlgorithm
from aligner import AlignmentOperation, QueryTargetAlignment
from shared_classes import TextChunk, LetterList
from text_loaders import TextLoader
from text_transformers import TextTransformer
from terminaltables import AsciiTable


def debugger_is_active() -> bool:
    """Return if the debugger is currently active"""
    gettrace = getattr(sys, "gettrace", lambda: None)
    return gettrace() is not None


class TextAlignmentTool:
    """
    The text alignment tool provides a context within which
    a custom and extensible pipeline of operations can be
    carried out. It tracks each operation in the pipeline
    process and facilitates user defined text alignment
    operations.
    """

    def __init__(self, query_loader: TextLoader, target_loader: TextLoader):
        if debugger_is_active():
            # In debugging mode, inject some root level utilities
            # to make it easire to visualize the text and chunks
            main_mod = sys.modules["__main__"]
            main_mod.__builtins__["dbg"] = DebugHelper()

        self.__query_loader: TextLoader = query_loader
        self.__target_loader: TextLoader = target_loader
        self.__operation_list: list[AlignmentOperation] = []
        self.__load_texts()

    def __load_texts(self):
        no_operation_query_transform = TextTransformer()
        no_operation_query_transform.load_input(
            self.__query_loader.output,
            self.__query_loader.text_chunk_indices,
            no_transform=True,
        )
        self.__operation_list.append(
            AlignmentOperation(no_operation_query_transform, None, None)
        )

        no_operation_target_transform = TextTransformer()
        no_operation_target_transform.load_input(
            self.__target_loader.output,
            self.__target_loader.text_chunk_indices,
            no_transform=True,
        )
        self.__operation_list.append(
            AlignmentOperation(None, no_operation_target_transform, None)
        )

    def query_text_transforms(self, query_transformers: list[TextTransformer]):
        """
        Apply any number of transformations to the query text
        in the order submitted.
        """

        for transformer in query_transformers:
            latest_query_transformer = self.__latest_query_transform
            transformer.load_input(
                latest_query_transformer.output,
                latest_query_transformer.text_chunk_indices,
            )
            self.__operation_list.append(AlignmentOperation(transformer, None, None))

    def query_text_transform(self, query_transformers: TextTransformer):
        """
        Apply one transformations to the query text.
        """

        self.query_text_transforms([query_transformers])

    def target_text_transforms(self, target_transformers: list[TextTransformer]):
        """
        Apply any number of transformations to the target text
        in the order submitted.
        """

        for transformer in target_transformers:
            latest_target_transformer = self.__latest_target_transform
            transformer.load_input(
                latest_target_transformer.output,
                latest_target_transformer.text_chunk_indices,
            )
            self.__operation_list.append(AlignmentOperation(None, transformer, None))

    def target_text_transform(self, target_transformer: TextTransformer):
        """
        Apply one transformations to the target text.
        """

        self.target_text_transforms([target_transformer])

    def align_texts(self, alignment_algorithms: list[AlignmentAlgorithm]):
        """
        Apply any number of alignment operations to the query and target
        texts in the order specified.
        """

        for alignment_algorithm in alignment_algorithms:
            latest_query_text = self.__latest_query_text
            latest_query_text_chunk_indices = self.__latest_query_text_chunk_indices
            latest_target_text = self.__latest_target_text
            latest_target_text_chunk_indices = self.__latest_target_text_chunk_indices
            alignment_algorithm.load_texts(
                latest_query_text,
                latest_target_text,
                latest_query_text_chunk_indices,
                latest_target_text_chunk_indices,
            )
            alignment_algorithm.align()
            self.__operation_list.append(
                AlignmentOperation(None, None, alignment_algorithm)
            )

    def align_text(self, alignment_algorithm: AlignmentAlgorithm):
        """
        Apply a single alignment operation to the query and target
        texts.
        """

        self.align_texts([alignment_algorithm])

    def collect_all_alignments(self) -> list[list[QueryTargetAlignment]]:
        """Based on the latest alignment, calculate all possible correspondances
        between every stage of the query and the target transformation
        pipeline.

        Returns:
            list[list[QueryTargetAlignment]]: this 2D list contains every possible
            alignment of the query and target. The 1st axis tracks the query text
            from latest to earliest. The 2nd axis tracks the target text from
            latest to earliest.
        """
        results: list[list[QueryTargetAlignment]] = []

        # Collect all possible query and target transforms
        query_transforms: list[AlignmentOperation] = [
            x
            for x in self.__operation_list
            if x.query_transformation is not None  # or x.alignment is not None
        ]
        target_transforms: list[AlignmentOperation] = [
            x
            for x in self.__operation_list
            if x.target_transformation is not None  # or x.alignment is not None
        ]

        latest_alignment: TextAlignments = self.latest_alignment._alignment
        for query_alignment_operation in reversed(query_transforms):
            # if alignment_operation.alignment is not None:
            #     latest_alignment = alignment_operation.alignment._alignment
            if (
                latest_alignment is not None
                and query_alignment_operation.query_transformation is not None
            ):
                stored_alignment: Union[TextAlignments, None] = None
                current_query_target_alignment: list[QueryTargetAlignment] = []
                current_query_text = (
                    query_alignment_operation.query_transformation.input
                )
                current_query_text_chunks = (
                    query_alignment_operation.query_transformation.input_text_chunk_indices
                )
                aligned_query_to_current_query: dict[int, list[int]] = {}
                for (
                    input_output_map_entry
                ) in query_alignment_operation.query_transformation.input_output_map:
                    if input_output_map_entry[1] not in aligned_query_to_current_query:
                        aligned_query_to_current_query[input_output_map_entry[1]] = []
                    aligned_query_to_current_query[input_output_map_entry[1]].append(
                        input_output_map_entry[0]
                    )

                # latest_alignment: Union[TextAlignments, None] = None
                for target_alignment_operation in reversed(target_transforms):
                    # if alignment_operation.alignment is not None:
                    #     latest_alignment = alignment_operation.alignment._alignment
                    if (
                        latest_alignment is not None
                        and target_alignment_operation.target_transformation is not None
                    ):
                        # latest_alignment = (
                        #     alignment_operation.target_transformation.apply_alignment(
                        #         latest_alignment
                        #     )
                        # )
                        current_target_text = (
                            target_alignment_operation.target_transformation.input
                        )
                        current_target_text_chunks = (
                            target_alignment_operation.target_transformation.input_text_chunk_indices
                        )
                        aligned_target_to_current_target: dict[int, list[int]] = {}
                        for (
                            input_output_map_entry
                        ) in (
                            target_alignment_operation.target_transformation.input_output_map
                        ):
                            if (
                                input_output_map_entry[1]
                                not in aligned_target_to_current_target
                            ):
                                aligned_target_to_current_target[
                                    input_output_map_entry[1]
                                ] = []
                            aligned_target_to_current_target[
                                input_output_map_entry[1]
                            ].append(input_output_map_entry[0])

                        new_alignment = TextAlignments()
                        for alignment in latest_alignment.alignments:
                            for query_idx in aligned_query_to_current_query.get(
                                alignment.query_idx, []
                            ):
                                for target_idx in aligned_target_to_current_target.get(
                                    alignment.target_idx, []
                                ):
                                    new_alignment.alignments.append(
                                        AlignmentKey(query_idx, target_idx)
                                    )
                        current_query_target_alignment.append(
                            QueryTargetAlignment(
                                new_alignment,
                                current_query_text,
                                current_target_text,
                                current_query_text_chunks,
                                current_target_text_chunks,
                                query_alignment_operation.query_transformation.__class__.__name__,
                                target_alignment_operation.target_transformation.__class__.__name__,
                            )
                        )

                        latest_alignment = new_alignment
                        if stored_alignment is None:
                            stored_alignment = latest_alignment

                if stored_alignment is not None:
                    latest_alignment = stored_alignment
                results.append(current_query_target_alignment)

        return results

    # def find_alignment_to_query(self) -> list[QueryTargetAlignment]:
    #     """
    #     Walk the latest alignment mapping back to the earliest
    #     form of the target text.
    #     """

    #     query_transforms: list[AlignmentOperation] = [
    #         x
    #         for x in self.__operation_list
    #         if x.query_transformation is not None or x.alignment is not None
    #     ]
    #     latest_target_text = self.__latest_target_text
    #     latest_alignment: Union[TextAlignments, None] = None
    #     alignment_history: list[QueryTargetAlignment] = []
    #     for alignment_operation in reversed(query_transforms):
    #         if alignment_operation.alignment is not None:
    #             latest_alignment = alignment_operation.alignment._alignment
    #         if (
    #             latest_alignment is not None
    #             and alignment_operation.query_transformation is not None
    #         ):
    #             latest_alignment = (
    #                 alignment_operation.query_transformation.apply_alignment(
    #                     latest_alignment
    #                 )
    #             )
    #             alignment_history.append(
    #                 QueryTargetAlignment(
    #                     latest_alignment,
    #                     alignment_operation.query_transformation.input,
    #                     alignment_operation.query_transformation.input_text_chunk_indices,
    #                     latest_target_text,
    #                 )
    #             )

    #     return alignment_history

    # def find_alignment_to_target(self) -> list[QueryTargetAlignment]:
    #     """
    #     Walk the latest alignment mapping back to the earliest
    #     form of the query text.
    #     """

    #     target_transforms: list[AlignmentOperation] = [
    #         x
    #         for x in self.__operation_list
    #         if x.target_transformation is not None or x.alignment is not None
    #     ]
    #     latest_query_text = self.__latest_query_text
    #     latest_alignment: Union[TextAlignments, None] = None
    #     alignment_history: list[QueryTargetAlignment] = []
    #     for alignment_operation in reversed(target_transforms):
    #         if alignment_operation.alignment is not None:
    #             latest_alignment = alignment_operation.alignment._alignment
    #         if (
    #             latest_alignment is not None
    #             and alignment_operation.target_transformation is not None
    #         ):
    #             latest_alignment = (
    #                 alignment_operation.target_transformation.apply_alignment(
    #                     latest_alignment
    #                 )
    #             )
    #             alignment_history.append(
    #                 QueryTargetAlignment(
    #                     latest_alignment,
    #                     alignment_operation.target_transformation.input,
    #                     alignment_operation.target_transformation.input_text_chunk_indices,
    #                     latest_query_text,
    #                 )
    #             )

    #     return alignment_history

    @property
    def latest_alignment(self) -> AlignmentAlgorithm:
        alignments = [
            x.alignment for x in self.__operation_list if x.alignment is not None
        ]

        if alignments:
            return alignments[-1]

        raise AlignmentException("No alignment found")

    @property
    def __latest_operation(self) -> AlignmentOperation:
        return self.__operation_list[-1]

    @property
    def __latest_query_operation_text_chunks(self) -> list[TextChunk]:
        transformer = [
            x for x in self.__operation_list if x.query_transformation is not None
        ][-1].query_transformation
        return transformer.text_chunk_indices if transformer is not None else []

    @property
    def __latest_target_operation_text_chunks(self) -> list[TextChunk]:
        transformer = [
            x for x in self.__operation_list if x.target_transformation is not None
        ][-1].target_transformation
        return transformer.text_chunk_indices if transformer is not None else []

    @property
    def __latest_query_text(self) -> LetterList:
        return self.__latest_query_transform.output

    @property
    def __latest_query_text_chunk_indices(self) -> list[TextChunk]:
        return (
            self.__latest_operation.alignment.output_query_text_chunk_indices
            if self.__latest_operation.alignment is not None
            else self.__latest_query_operation_text_chunks
        )

    @property
    def __latest_query_transform(self) -> TextTransformer:
        return self.__query_text_operations[-1]

    @property
    def __query_text_operations(self) -> list[TextTransformer]:
        return [
            x.query_transformation
            for x in self.__operation_list
            if x.query_transformation is not None
        ]

    @property
    def __latest_target_text(self) -> LetterList:
        return self.__latest_target_transform.output

    @property
    def __latest_target_text_chunk_indices(self) -> list[TextChunk]:
        return (
            self.__latest_operation.alignment.output_target_text_chunk_indices
            if self.__latest_operation.alignment is not None
            else self.__latest_target_operation_text_chunks
        )

    @property
    def __latest_target_transform(self) -> TextTransformer:
        return self.__target_text_operations[-1]

    @property
    def __target_text_operations(self) -> list[TextTransformer]:
        return [
            x.target_transformation
            for x in self.__operation_list
            if x.target_transformation is not None
        ]

    def check(self):
        pass


def get_text_for_range(text: LetterList, start_idx: int, end_idx: int) -> str:
    return "".join([chr(x) for x in text[start_idx : end_idx + 1]])


def get_text_chunk_string(text: LetterList, chunk: TextChunk) -> str:
    return "".join([chr(x) for x in text[chunk.indices]])


def get_text_chunks_string(
    text: LetterList, chunks: Sequence[TextChunk], joiner: str = "\n"
) -> str:
    return joiner.join([get_text_chunk_string(text, x) for x in chunks])


def compare_parallel_text_chunks(
    query: LetterList,
    query_chunks: Sequence[TextChunk],
    target: LetterList,
    target_chunks: Sequence[TextChunk],
    rtl=False,
) -> str:
    parallelized_text: list[list[str]] = [
        [
            get_text_chunk_string(query, x)[:: -1 if rtl else 1],
            get_text_chunk_string(target, y)[:: -1 if rtl else 1],
        ]
        for x, y in zip(query_chunks, target_chunks)
    ]
    parallelized_text.insert(0, ["query", "target"])
    table = AsciiTable(parallelized_text)
    table.inner_column_border = True
    table.justify_columns[0] = "center"
    table.justify_columns[1] = "center"
    return table.table


def print_parallel_text_chunks(
    query: LetterList,
    query_chunks: Sequence[TextChunk],
    target: LetterList,
    target_chunks: Sequence[TextChunk],
    rtl=False,
):
    comparison = compare_parallel_text_chunks(
        query, query_chunks, target, target_chunks, rtl
    )
    print(comparison)


class DebugHelper:
    """
    The Debug helper provides several convenience methods for inspecting
    the text inside the alignment tool.
    """

    def __init__(self):
        pass

    @staticmethod
    def display_text_region(text: LetterList, start_idx: int, end_idx: int) -> str:
        """
        Print out a human readable representation of the letter list
        for the section between the start_idx and end_idx.
        """

        text_chunk = "".join([chr(x) for x in text[start_idx:end_idx]])
        print(text_chunk)
        return text_chunk

    @classmethod
    def display_text_chunk(cls, text: LetterList, chunk: TextChunk) -> str:
        """
        Print out a human readable representation of the letter list
        for the specified text chunk.
        """
        text_chunk = "".join([chr(x) for x in text[chunk.indices]])
        print(text_chunk)
        return text_chunk

    @classmethod
    def display_text_chunks(
        cls, text: LetterList, chunks: list[TextChunk]
    ) -> list[str]:
        """
        Print out a human readable representation of the letter list
        for the specified text chunks.
        """

        return [cls.display_text_chunk(text, x) for x in chunks]

    @classmethod
    def display_text(cls, text: LetterList) -> str:
        """
        Print out a human readable representation of the entire
        letter list.
        """

        return cls.display_text_region(text, 0, len(text) - 1)



from .text_loader import TextLoader
from .alto_xml_loader import AltoXMLTextLoader
from .pgp_xml_loader import PgpXmlTeiTextLoader
from .string_text_loader import StringTextLoader
from .newline_separated_text_loader import NewlineSeparatedTextLoader


import logging
import swalign
from .alignment_algorithm import AlignmentAlgorithm, TextAlignments, AlignmentKey
from ..shared_classes import TextChunk


class LocalAlignmentAlgorithm(AlignmentAlgorithm):
    def __init__(self):
        super().__init__()

    def align(self) -> TextAlignments:
        """
        Perform a local alignment of the query and target.  Primarily
        this is useful for finding a small chunk of query text within
        a large chunk of target text. It will probably not be able to
        provide a good  alignment on the letter level, but it should be
        quite close on the chunk level. The output of
        `output_query_text_chunk_indices` and
        `output_target_text_chunk_indices` can be used to find a more
        precise global alignment.
        """

        if len(self._input_query_text_chunk_indices) != len(
            self._input_target_text_chunk_indices
        ):
            logging.warning(
                f"Since the query has {len(self._input_query_text_chunk_indices)} text chunks and the target has {len(self._input_target_text_chunk_indices)}, the alignment will probably not work as intended."
            )

        output_query_chunks: list[TextChunk] = []
        output_target_chunks: list[TextChunk] = []
        query_start_end = [
            (x.start_idx, x.end_idx) for x in self._input_query_text_chunk_indices
        ]
        target_start_end = [
            (x.start_idx, x.end_idx) for x in self._input_target_text_chunk_indices
        ]
        for (query_start_idx, query_end_idx), (target_start_idx, target_end_idx) in zip(
            query_start_end, target_start_end
        ):
            alignments = local_alignment(
                self._query[query_start_idx : query_end_idx + 1],
                self._target[target_start_idx : target_end_idx + 1],
            )
            output_query_chunks.append(TextChunk(alignments[0], alignments[1]))
            output_target_chunks.append(TextChunk(alignments[2], alignments[3]))

        # Note that these alignments will probably not be very good
        text_alignments: list[AlignmentKey] = []
        for query_chunk, target_chunk in zip(output_query_chunks, output_target_chunks):
            max_target_idx = target_chunk.end_idx + 1
            for idx in range(query_chunk.end_idx - query_chunk.start_idx + 1):
                query_idx = query_chunk.start_idx + idx
                target_idx = (
                    target_chunk.start_idx + idx
                    if target_chunk.start_idx + idx < max_target_idx
                    else max_target_idx
                )
                text_alignments.append(AlignmentKey(query_idx, target_idx))

        # Save alignments to self object
        self._output_query_text_chunk_indices = output_query_chunks
        self._output_target_text_chunk_indices = output_target_chunks
        self._alignment.alignments = text_alignments

        return super().align()


def local_alignment(query_text, target_text) -> tuple[int, int, int, int]:
    # choose your own values here… 2 and -1 are common.
    match = 2
    mismatch = -1
    scoring = swalign.NucleotideScoringMatrix(match, mismatch)

    sw = swalign.LocalAlignment(scoring)  # you can also choose gap penalties, etc...
    target_transcription = " ".join([chr(x) for x in target_text])
    query_transcription = " ".join([chr(x) for x in query_text])
    alignment = sw.align(target_transcription, query_transcription)
    start_idx, end_idx, query_start, query_end = (
        alignment.r_pos,
        alignment.r_end,
        alignment.q_pos,
        alignment.q_end,
    )
    alignment.dump()
    return start_idx, end_idx, query_start, query_end



from alignment_algorithms import (
    AlignmentAlgorithm,
    TextAlignments,
    AlignmentKey,
    AlignmentException,
    GlobalAlignmentAlgorithm,
    LocalAlignmentAlgorithm,
    RoughAlignmentAlgorithm,
    ChunkAlignmentAlgorithm,
)

from alignment_tool import (
    TextAlignmentException,
    AlignedIndices,
    AlignmentTextDataObject,
    AlignmentOperation,
    TextAlignmentTool,
    get_text_for_range,
    get_text_chunk_string,
    get_text_chunks_string,
    compare_parallel_text_chunks,
    print_parallel_text_chunks,
)

from find_wordlist_for_alignment import find_wordlist_for_alignment

from shared_classes import LetterList, TextChunk

from text_loaders import (
    TextLoader,
    AltoXMLTextLoader,
    PgpXmlTeiTextLoader,
    StringTextLoader,
    NewlineSeparatedTextLoader,
)

from text_transformers import (
    TextTransformer,
    TransformerError,
    RemoveCharacter,
    RemoveCharacterTransformer,
    CharacterSubstitution,
    SubstituteCharacterTransformer,
    MultipleCharacterSubstitution,
    SubstituteMultipleCharactersTransformer,
    BracketingPair,
    ConsistentBracketingTransformer,
)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['text_alignment',
 'text_alignment.alignment_algorithms',
 'text_alignment.alignment_tool',
 'text_alignment.find_wordlist_for_alignment',
 'text_alignment.shared_classes',
 'text_alignment.text_loaders',
 'text_alignment.text_transformers']

package_data = \
{'': ['*']}

install_requires = \
['bidict>=0.21.3,<0.22.0',
 'colored>=1.4.2,<2.0.0',
 'cursive-re>=0.0.4,<0.0.5',
 'dotmap>=1.3.24,<2.0.0',
 'edlib>=1.3.9,<2.0.0',
 'fuzzywuzzy>=0.18.0,<0.19.0',
 'lxml>=4.6.3,<5.0.0',
 'numpy>=1.17,<2.0',
 'pandas>=1.3.3,<2.0.0',
 'swalign>=0.3.6,<0.4.0',
 'terminaltables>=3.1.0,<4.0.0',
 'tqdm>=4.62.2,<5.0.0',
 'typing>=3.6.0,<4.0.0']

setup_kwargs = {
    'name': 'text-alignment-tool',
    'version': '0.2.3',
    'description': 'A pipeline tool for performing customized text alignment procedures',
    'long_description': None,
    'author': 'Bronson Brown-deVost',
    'author_email': 'bronsonbdevost@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from shared_classes import LetterList, TextChunk
from text_transformers import TextTransformer
from itertools import zip_longest
import numpy as np


class MultipleCharacterSubstitution:
    def __init__(self, input_characters: str, output_characters: str) -> None:
        if input_characters == output_characters:
            raise Exception(
                "Input and output strings are the same, no passthroughs allowed."
            )
        if len(input_characters) == 0:
            raise Exception("Input characters is empty.")
        self.input_characters = np.array(
            [ord(x) for x in self.verify_characters(input_characters)]
        )
        self.output_characters = np.array(
            [ord(x) for x in self.verify_characters(output_characters)]
        )

    @staticmethod
    def verify_characters(char: str):
        if not isinstance(char, (str)):
            raise TypeError(f"Only str types are accepted: {char} is {type(char)} ")
        return char


class SubstituteMultipleCharactersTransformer(TextTransformer):
    def __init__(self, substitutions_list: list[MultipleCharacterSubstitution]):
        """Perform a transform of a single character to another single character

        Args:
            substitution_list (list[CharacterSubstitution]): a list of input/output character substitutions
        """
        self.substitutions_list = substitutions_list
        super().__init__()

    # def load_input(self, input: LetterList, text_chunk_indices: list[TextChunk]):
    #     super().load_input(input, text_chunk_indices, False)

    def transform(self) -> LetterList:
        if np.any(self._output):
            return self._output

        self.__transform()

        return super().transform()

    def __transform(self):
        replacement_transforms = [x.input_characters for x in self.substitutions_list]
        output = np.array([], dtype=np.uint32)
        input_output_map: list[tuple[int, int]] = []
        output_chunks: list[TextChunk] = []
        for chunk in self._input_text_chunk_indices:
            output_chunk = TextChunk([], chunk.name)
            output_chunk_start_idx = len(output)
            chunk_text = self.input[chunk.indices]
            altered_chunk_text = chunk_text.copy()
            chunk_in_out_map = [(x, x) for x in range(chunk_text.size)]
            matches_found = True
            while matches_found:
                matches_found = False
                for idx, search_text in enumerate(replacement_transforms):
                    matches = find_sequence_in_array(search_text, altered_chunk_text)
                    if not matches:
                        continue
                    matches_found = True
                    replacement_text = self.substitutions_list[idx].output_characters
                    replacement_offset = replacement_text.size - search_text.size
                    altered_text_offset = chunk_text.size - altered_chunk_text.size
                    altered_chunk_text = np.concatenate(
                        (
                            altered_chunk_text[: matches[0][0]],
                            replacement_text,
                            altered_chunk_text[matches[0][0] + search_text.size :],
                        )
                    ).astype(np.uint32)
                    max_middle_altered_chunk = (matches[0][0] + search_text.size) % (
                        matches[0][0] + replacement_text.size
                    )
                    new_starting_chunk_map = chunk_in_out_map[0 : matches[0][0]]
                    new_middle_chunk_map = [
                        (i + altered_text_offset, o)
                        for o, i in zip_longest(
                            range(
                                matches[0][0],
                                matches[0][0] + replacement_text.size,
                            ),
                            range(matches[0][0], max_middle_altered_chunk),
                            fillvalue=max_middle_altered_chunk - 1,
                        )
                    ]
                    new_ending_chunk_map = [
                        (x[0], x[1] + replacement_offset)
                        for x in chunk_in_out_map[matches[0][0] + search_text.size :]
                    ]
                    chunk_in_out_map = (
                        new_starting_chunk_map
                        + new_middle_chunk_map
                        + new_ending_chunk_map
                    )

            output = np.concatenate((output, altered_chunk_text))
            input_output_map = input_output_map + [
                (chunk.start_idx + x[0], output_chunk_start_idx + x[1])
                for x in chunk_in_out_map
            ]
            if output_chunk_start_idx <= len(output) - 1:
                output_chunks.append(
                    TextChunk(output_chunk_start_idx, len(output) - 1, chunk.name)
                )

        self._input_output_map = input_output_map
        self._output = output
        self._text_chunk_indices = output_chunks


def find_sequence_in_array(seq: np.ndarray, arr: np.ndarray) -> list[int]:
    """Find sequence in an array using NumPy only.
        See: https://stackoverflow.com/questions/36522220/searching-a-sequence-in-a-numpy-array

    Parameters
    ----------
    arr    : input 1D array
    seq    : input 1D array

    Output
    ------
    Output : 1D list of starting indices in the input array that satisfy the
    matching of input sequence in the input array.
    In case of no match, an empty list is returned.
    """

    # Store sizes of input array and sequence
    Na, Nseq = arr.size, seq.size
    if Na == 0 or Nseq == 0:
        return []

    # Range of sequence
    r_seq = np.arange(Nseq)

    # Create a 2D array of sliding indices across the entire length of input array.
    # Match up with the input sequence & get the matching starting indices.
    M = (arr[np.arange(Na - Nseq + 1)[:, None] + r_seq] == seq).all(1)

    # Get the range of those indices as final output
    if M.any() > 0:
        return np.where(M)
    else:
        return []



import edlib
from text_alignment_tool.src.shared_classes import LetterList, TextChunk
from text_alignment_tool.src.alignment_algorithms import (
    AlignmentAlgorithm,
    TextAlignments,
    AlignmentKey,
)
from statistics import mean
from fuzzywuzzy import fuzz
import swalign


class ChunkAlignmentAlgorithm(AlignmentAlgorithm):
    """Aligns two texts based upon the chunking of the texts.
    It aims to match corresponding corresponding lines based
    upon greatest similarity, without making any deletions
    from the target text.
    """

    def __init__(self):
        super().__init__()

    def align(self) -> TextAlignments:
        if self._alignment.alignments:
            return self._alignment

        linewise_matches = comprehensive_linewise_matching(
            self._query,
            self._target,
            self._input_query_text_chunk_indices,
            self._input_target_text_chunk_indices,
        )

        # Find best matches for lines
        alignment = TextAlignments()
        linewise_dict: dict[int, list[int]] = {}
        for linewise_match in linewise_matches:
            if linewise_match[0] not in linewise_dict:
                linewise_dict[linewise_match[0]] = []
            linewise_dict[linewise_match[0]].append(linewise_match[1])

        # Calculate the actual corresponding chunks of the lines
        query_target_text_chunk_alignment: list[tuple[TextChunk, TextChunk]] = []
        for linewise_match_key in linewise_dict.keys():
            query_text_chunk_indices = self._input_query_text_chunk_indices[
                linewise_match_key
            ]
            if len(linewise_dict[linewise_match_key]) == 1:
                query_target_text_chunk_alignment.append(
                    (
                        query_text_chunk_indices,
                        self._input_target_text_chunk_indices[
                            linewise_dict[linewise_match_key][0]
                        ],
                    )
                )
                continue

            query_text = self._query[
                query_text_chunk_indices.start_idx : query_text_chunk_indices.end_idx
                + 1
            ]
            for target_line_index in sorted(
                linewise_dict[linewise_match_key],
                key=lambda k: self._input_target_text_chunk_indices[k].end_idx
                - self._input_target_text_chunk_indices[k].start_idx,
                reverse=True,
            ):
                target_text_chunk_indices = self._input_target_text_chunk_indices[
                    target_line_index
                ]
                target_text = self._target[
                    target_text_chunk_indices.start_idx : target_text_chunk_indices.end_idx
                    + 1
                ]
                (
                    target_start_idx,
                    target_end_idx,
                    query_start,
                    query_end,
                ) = perform_local_chunk_alignment(query_text, target_text)
                query_target_text_chunk_alignment.append(
                    (
                        TextChunk(
                            query_text_chunk_indices.start_idx + query_start,
                            query_text_chunk_indices.end_idx + query_end + 1,
                        ),
                        TextChunk(
                            target_text_chunk_indices.start_idx + target_start_idx,
                            target_text_chunk_indices.end_idx + target_end_idx,
                        ),
                    )
                )
                # Nullify the matched sequence
                section_length = len(query_text[query_start : query_end + 1])
                query_text[query_start : query_end + 1] = [1] * section_length

        # Build also the full alignment (based upon the chunk alignment)
        for chunk_alignment in query_target_text_chunk_alignment:
            query = self._query[
                chunk_alignment[0].start_idx : chunk_alignment[0].end_idx + 1
            ]
            target = self._target[
                chunk_alignment[1].start_idx : chunk_alignment[1].end_idx + 1
            ]
            if query.size == 0 or target.size == 0:
                continue

            best_alignment = perform_global_alignment(
                query,
                target,
                chunk_alignment[0].start_idx,
                chunk_alignment[1].start_idx,
            )
            alignment.alignments = alignment.alignments + best_alignment

        # self.__query_target_text_chunk_alignment = query_target_text_chunk_alignment
        self._output_query_text_chunk_indices = [
            x[0] for x in query_target_text_chunk_alignment
        ]
        self._output_target_text_chunk_indices = [
            x[1] for x in query_target_text_chunk_alignment
        ]
        self._alignment = alignment
        return self._alignment


def comprehensive_linewise_matching(
    query_text: LetterList,
    target_text: LetterList,
    query_text_chunk_bounds: list[TextChunk],
    target_text_chunk_bounds: list[TextChunk],
) -> list[tuple[int, int]]:
    query_text_chunks = [
        "".join(chr(y) for y in query_text[x.start_idx : x.end_idx + 1])
        for x in query_text_chunk_bounds
    ]
    mean_query_length = mean([len(x) for x in query_text_chunks])
    target_text_chunks = [
        "".join(chr(y) for y in target_text[x.start_idx : x.end_idx + 1])
        for x in target_text_chunk_bounds
    ]
    mean_target_length = mean([len(x) for x in target_text_chunks])

    scored_matches: list[tuple[int, int, int]] = []
    for target_index, target_text_chunk in enumerate(target_text_chunks):
        for query_index, query_text_chunk in enumerate(query_text_chunks):
            ratio = (
                fuzz.partial_ratio
                if len(query_text_chunk) < (mean_query_length / 2)
                and len(target_text_chunk) < (mean_target_length / 2)
                else fuzz.ratio
            )
            pair_distance = ratio(query_text_chunk, target_text_chunk)
            scored_matches.append((pair_distance, query_index, target_index))

    best_matches: list[tuple[int, int, int]] = []
    sorted_scored_matches = sorted(scored_matches, key=lambda x: x[0], reverse=True)
    remaining_matches: list[tuple[int, int, int]] = sorted_scored_matches.copy()
    for match in sorted_scored_matches:
        if match not in remaining_matches:
            continue

        best_matches.append(match)
        remaining_matches = list(
            filter(
                lambda x: x[1] != match[1] and x[2] != match[2],
                remaining_matches,
            )
        )

    return [(x[1], x[2]) for x in sorted(best_matches, key=lambda x: x[1])]


def perform_global_alignment(
    query_text: LetterList,
    target_text: LetterList,
    query_offset: int,
    target_offset: int,
) -> list[AlignmentKey]:
    alignments: list[AlignmentKey] = []
    query = "".join([chr(x) for x in query_text])
    target = "".join([chr(x) for x in target_text])

    result = edlib.align(query, target, task="path", mode="HW")
    nice = edlib.getNiceAlignment(result, query, target)
    query_count_idx = query_offset
    target_count_idx = target_offset + result["locations"][0][0]
    for idx in range(len(nice["query_aligned"])):
        alignments.append(AlignmentKey(query_count_idx, target_count_idx))
        if nice["query_aligned"][idx] != "-":
            query_count_idx += 1
        if nice["target_aligned"][idx] != "-":
            target_count_idx += 1

    return alignments


def perform_local_chunk_alignment(
    query_text: LetterList,
    target_text: LetterList,
) -> tuple[int, int, int, int]:
    # choose your own values here… 2 and -1 are common.
    match = 2
    mismatch = -1
    scoring = swalign.NucleotideScoringMatrix(match, mismatch)

    sw = swalign.LocalAlignment(scoring)  # you can also choose gap penalties, etc...
    target_transcription = "".join([chr(x) for x in target_text])
    query_transcription = "".join([chr(x) for x in query_text])
    alignment = sw.align(target_transcription, query_transcription)
    target_start_idx, target_end_idx, query_start, query_end = (
        alignment.r_pos,
        alignment.r_end,
        alignment.q_pos,
        alignment.q_end,
    )
    alignment.dump()

    return target_start_idx, target_end_idx, query_start, query_end



from alignment_algorithms import TextAlignments, AlignmentKey
from shared_classes import LetterList, TextChunk
import numpy as np


class TransformerError(Exception):
    pass


class TextTransformer:
    """
    Instructions for a specific text transformation operation.
    """

    def __init__(self):
        self._input: LetterList = np.array([], dtype=np.uint32)
        self._output: LetterList = np.array([], dtype=np.uint32)
        self._input_output_map: list[tuple[int, int]] = []
        self._input_text_chunk_indices: list[TextChunk] = []
        self._text_chunk_indices: list[TextChunk] = []

    def load_input(
        self,
        input: LetterList,
        text_chunk_indices: list[TextChunk],
        no_transform: bool = False,
    ):
        if self._input.size > 0:
            return

        self._input = input.astype(np.uint32)
        self._input_text_chunk_indices = text_chunk_indices

        if no_transform:
            self._output = self._input
            self._input_output_map = [(idx, idx) for idx in range(len(self._input))]
            self._text_chunk_indices = self._input_text_chunk_indices

        self.transform()

    def transform(self) -> LetterList:
        if self._output.size > 0:
            return self._output

        raise TransformerError("Transform action resulted in empty output")

    def apply_alignment(self, alignment_pairs: TextAlignments) -> TextAlignments:
        # Note: I think this will only work when the input length is >= the output length
        input_to_aligned_text_map = TextAlignments()
        alignment_pairs_dict: dict[int, list[int]] = {}

        for alignment in alignment_pairs.alignments:
            if alignment.query_idx not in alignment_pairs_dict:
                alignment_pairs_dict[alignment.query_idx] = []

            alignment_pairs_dict[alignment.query_idx].append(alignment.target_idx)

        for input_idx, output_idx in self.input_output_map:
            if output_idx in alignment_pairs_dict:
                for align_idx in alignment_pairs_dict[output_idx]:
                    input_to_aligned_text_map.alignments.append(
                        AlignmentKey(input_idx, align_idx)
                    )

        return input_to_aligned_text_map

    @property
    def full_text(self) -> str:
        return "".join([chr(x) for x in self.output])

    @property
    def chunked_text(self) -> list[tuple[str, str]]:
        return [
            (y.name, "".join([chr(self.output[x]) for x in y.indices]))
            for y in self.text_chunk_indices
        ]

    @property
    def input(self) -> LetterList:
        if np.any(self._input):
            return self._input

        raise TransformerError("Input is empty")

    @property
    def output(self) -> LetterList:
        if not np.any(self._output):
            _ = self.transform()

        if np.any(self._output.size):
            return self._output

        raise TransformerError("Output is empty")

    @property
    def input_output_map(self) -> list[tuple[int, int]]:
        if not self._input_output_map:
            _ = self.output

        if self._input_output_map:
            return self._input_output_map

        raise TransformerError("Input output map is empty")

    @property
    def input_text_chunk_indices(self) -> list[TextChunk]:
        return self._input_text_chunk_indices

    @property
    def text_chunk_indices(self) -> list[TextChunk]:
        if not self._text_chunk_indices:
            self._text_chunk_indices.append(TextChunk(0, self.output.size))

        if self._text_chunk_indices:
            return self._text_chunk_indices

        raise TransformerError(
            "No text chunk was created (this is a development error)"
        )



from lxml import etree
from pathlib import Path
from typing import NamedTuple
from io import BufferedReader
import cursive_re

from shared_classes import TextChunk, LetterList
from text_loader import TextLoader
import numpy as np


class PgpXmlTeiInputPosition(NamedTuple):
    filename: str
    line_number: str
    word_index: int


class PgpXmlTeiAlignmentMap(NamedTuple):
    input_position: PgpXmlTeiInputPosition
    output_idx: int


class PgpXmlTeiTextLoader(TextLoader):
    def __init__(self, file_paths: list[Path]):
        super().__init__(file_paths)
        self._input_output_map: list[PgpXmlTeiAlignmentMap] = []

    def _load(self) -> LetterList:
        self.parse_pgp_tei_files()
        return super()._load()

    @staticmethod
    def close_open_brackets(text: str):
        if not text:
            return text

        unclosed_line_beginning = (
            cursive_re.beginning_of_line()
            + cursive_re.zero_or_more(cursive_re.none_of("["))
            + cursive_re.text("]")
        )
        unclosed_line_beginning_re = cursive_re.compile(unclosed_line_beginning)
        if unclosed_line_beginning_re.match(text):
            text = "[" + text

        if unclosed_line_beginning_re.match(
            text[::-1].replace("[", ";").replace("]", "[").replace(";", "]")
        ):
            text = text + "]"

        return text

    def parse_pgp_tei_files(self):
        events = ("start", "end")
        needed_elements = {
            "tei_header": "{http://www.tei-c.org/ns/1.0}teiHeader",
            "text": "{http://www.tei-c.org/ns/1.0}text",
            "body": "{http://www.tei-c.org/ns/1.0}body",
            "div": "{http://www.tei-c.org/ns/1.0}div",
            "l": "{http://www.tei-c.org/ns/1.0}l",
        }
        for file in self._file_paths:
            with open(file, "rb") as xml_data:
                self.parse_tei(xml_data, file.name, events, needed_elements)

    def parse_tei(
        self,
        xml_data: BufferedReader,
        filename: str,
        events: tuple[str, str],
        needed_elements: dict,
    ):
        letter_list: list[int] = []
        block = -1
        line = -1
        response = {"part": []}
        token_idx_count = 0
        text_chunk_indices: list[TextChunk] = []

        for action, element in etree.iterparse(
            xml_data, events=events, tag=needed_elements.values()
        ):
            if action == "start":
                if element.tag == needed_elements["tei_header"]:
                    response["header_metadata"] = etree.tostring(element).decode()

                elif element.tag == needed_elements["div"]:
                    response["part"].append({"block": [{"lines": []}]})
                    block += 1
                    line = -1

                elif element.tag == needed_elements["l"]:
                    line_number = element.attrib.get("n", "")
                    text = self.close_open_brackets(element.text)
                    if text:
                        chunk_start_idx = len(self._output)
                        line += 1
                        for idx, token in enumerate(text):
                            token_idx_count += 1
                            letter_list.append(ord(token))
                            self._input_output_map.append(
                                PgpXmlTeiAlignmentMap(
                                    PgpXmlTeiInputPosition(filename, line_number, idx),
                                    len(self._output) - 1,
                                )
                            )
                        chunk_end_idx = len(self._output) - 1
                        if chunk_start_idx < chunk_end_idx:
                            text_chunk_indices.append(
                                TextChunk(chunk_start_idx, chunk_end_idx)
                            )

            if action == "end":
                if element.tag in [
                    needed_elements["tei_header"],
                    needed_elements["div"],
                ]:
                    element.clear()

        self._output: LetterList = np.array(letter_list)
        self._text_chunk_indices = text_chunk_indices
